package org.knime.build.actors;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.net.URL;
import java.util.Map;
import java.util.jar.Attributes;
import java.util.jar.Manifest;
import java.util.regex.Pattern;

import org.eclipse.buckminster.core.TargetPlatform;
import org.eclipse.buckminster.core.actor.AbstractActor;
import org.eclipse.buckminster.core.actor.IActionContext;
import org.eclipse.buckminster.core.cspec.PathGroup;
import org.eclipse.buckminster.core.cspec.model.Action;
import org.eclipse.buckminster.core.cspec.model.ComponentIdentifier;
import org.eclipse.buckminster.core.cspec.model.TopLevelAttribute;
import org.eclipse.buckminster.core.helpers.FileUtils;
import org.eclipse.buckminster.core.metadata.model.IModelCache;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.debug.core.ILaunchManager;
import org.eclipse.debug.core.model.IProcess;
import org.eclipse.jdt.launching.IJavaLaunchConfigurationConstants;
import org.osgi.framework.Bundle;

public class DocBuilderActor extends AbstractActor {
    private static final FilenameFilter FILTER = new FilenameFilter() {
        @Override
        public boolean accept(final File dir, final String name) {
            return !".svn".equals(name);
        }
    };

    @Override
    protected IStatus internalPerform(final IActionContext ctx,
            final IProgressMonitor monitor) throws CoreException {
        // first modify the MANIFEST.MF
        for (Map.Entry<String, PathGroup[]> e : ctx.getNamedPathGroupArrays()
                .entrySet()) {
            if (e.getKey().endsWith(".manifest")) {
                for (PathGroup pg : e.getValue()) {
                    IPath base = pg.getBase();
                    for (IPath path : pg.getPaths()) {
                        if ("MANIFEST.MF".equals(path.lastSegment())) {
                            IPath manifest = base.append(path);
                            try {
                                updateManifest(manifest.toFile());
                            } catch (Exception ex) {
                                return new Status(IStatus.ERROR,
                                        "org.knime.build",
                                        "Could not change MANIFEST.MF", ex);
                            }
                        }
                    }
                }
            }
        }

        // then create help files
        // The order here is important because otherwise the up-to-date check
        // does not work, because the MANIFEST is also a pre-requisite of
        // this actor.
        return createHelpFiles(ctx, monitor);
    }

    private IStatus createHelpFiles(final IActionContext ctx,
            final IProgressMonitor monitor) throws CoreException,
            DebugException {
        Bundle myself = Platform.getBundle("org.knime.build");
        URL launchConfigUrl = myself.getEntry("/DocBuilder.launch");

        // look up or create a temporary project for storing
        // the launch configuration
        IProject tempProject =
                ResourcesPlugin.getWorkspace().getRoot()
                        .getProject(".buckminster");
        if (!tempProject.exists()) {
            IProjectDescription desc =
                    ResourcesPlugin.getWorkspace().newProjectDescription(
                            ".buckminster");
            tempProject.create(desc, IResource.HIDDEN, null);
        }
        if (!tempProject.isOpen()) {
            tempProject.open(null);
        }

        // copy docbuilder launch config from plugin to temp project
        IFile lcFile = tempProject.getFile("KNIME_DocBuilder.launch");
        if (lcFile.exists()) {
            lcFile.delete(true, null);
        }

        try {
            lcFile.create(launchConfigUrl.openStream(), true, null);
        } catch (IOException ex) {
            return new Status(IStatus.ERROR, "org.knime.build",
                    "Could not copy 'DocBuilder.launch' to temp file", ex);
        }

        ILaunchManager lm = DebugPlugin.getDefault().getLaunchManager();
        ILaunchConfiguration launchConfigTemplate =
                lm.getLaunchConfiguration(lcFile);
        ILaunchConfigurationWorkingCopy launchConfig =
                launchConfigTemplate.copy("CustomDocBuilder");

        // workaround for pre-2.4 builds, the docbuilder application was in
        // a different plugin
        boolean helpViewExists = false;
        for (ComponentIdentifier i : TargetPlatform.getInstance().getComponents()) {
            if ("org.knime.workbench.helpview".equals(i.getName())) {
                helpViewExists = true;
                break;
            }
        }

        if (helpViewExists || ResourcesPlugin.getWorkspace().getRoot()
                .getProject("org.knime.workbench.helpview").exists()) {
            launchConfig.setAttribute("application",
                    "org.knime.workbench.helpview.KNIME_Documentation_Builder");
        }

        IPath outputDir =
                ctx.getAction().getExpandedBase(
                        ctx.getAction().getProductBase(), ctx.getProperties());
        outputDir.toFile().mkdirs();

        Object pluginPattern = getActorProperty("pluginPattern");
        if (pluginPattern == null) {
            pluginPattern =
                    Pattern.quote(ctx.getCSpec().getComponentIdentifier()
                            .getName());
        }
        String programArguments =
                launchConfig
                        .getAttribute(
                                IJavaLaunchConfigurationConstants.ATTR_PROGRAM_ARGUMENTS,
                                "");
        programArguments += " -pluginPattern \"" + pluginPattern + "\"";
        programArguments += " -destination \"" + outputDir + "\"";

        launchConfig.setAttribute(
                IJavaLaunchConfigurationConstants.ATTR_PROGRAM_ARGUMENTS,
                programArguments);

        try {
            ILaunch il =
                    launchConfig.launch(ILaunchManager.RUN_MODE, monitor,
                            false, false);
            while (!il.isTerminated()) {
                Thread.sleep(250);
            }
            for (IProcess p : il.getProcesses()) {
                if (p.getExitValue() != 0) {
                    return new Status(IStatus.ERROR, "org.knime.build",
                            "DocBuilder finished abnormally");
                }
            }
        } catch (InterruptedException e) {
            return Status.CANCEL_STATUS;
        } catch (CoreException ex) {
            ex.printStackTrace();
            throw ex;
        } finally {
            launchConfig.delete();
            lcFile.delete(true, null);
        }

        // MANIFEST.MF is taken from the prerequisites
        File metaInf = new File(outputDir.toFile(), "META-INF");
        FileUtils.deleteRecursive(metaInf, monitor);

        File pluginXml = new File(outputDir.toFile(), "plugin.xml");
        if (!pluginXml.exists()) {
            return new Status(
                    IStatus.ERROR,
                    "org.knime.build",
                    "plugin.xml not found after help files generation. Most "
                            + "likely this plugin does not contribute any nodes, "
                            + "metanodes, or categories, but help file generation is "
                            + "enabled. Please disable help file generation for this "
                            + "plugin");
        }

        return Status.OK_STATUS;
    }

    /**
     * Adds two necessary required bundles to the <tt>MANIFEST.MF</tt>.
     *
     * @param manifestFile the manifest file that should be changed
     * @throws IOException if an I/O error occurs
     */
    private void updateManifest(final File manifestFile) throws IOException {
        Manifest manifest = new Manifest(new FileInputStream(manifestFile));
        Attributes attrs = manifest.getMainAttributes();
        for (Map.Entry<Object, Object> o : manifest.getMainAttributes()
                .entrySet()) {
            if (o.getKey().equals(new Attributes.Name("Require-Bundle"))) {
                if (!((String)o.getValue()).contains("org.eclipse.help")) {
                    attrs.putValue(o.getKey().toString(), o.getValue() + ","
                            + "org.eclipse.help");
                }
                if (!((String)o.getValue()).contains("org.eclipse.help.ui")) {
                    attrs.putValue(o.getKey().toString(), o.getValue() + ","
                            + "org.eclipse.help.ui");
                }
                break;
            }
        }

        FileOutputStream outStream = new FileOutputStream(manifestFile);
        manifest.write(outStream);
        outStream.close();
    }

    @Override
    public boolean isUpToDate(final Action act, final IModelCache ctx,
            final long prerequisiteAge, final long oldestTarget)
            throws CoreException {
        long oldestProduct = getFirstModified(act, ctx);

        long youngestPrerequisite =
                getLastModified(act.getPrerequisiteGroup(), ctx, oldestProduct);

        return oldestProduct > youngestPrerequisite;
    }

    private long getFirstModified(final TopLevelAttribute att,
            final IModelCache ctx) throws CoreException {
        PathGroup[] pqs = att.getPathGroups(ctx, null);
        int idx = pqs.length;
        if (idx == 0) {
            return 0L;
        }

        long oldest = Long.MAX_VALUE;
        for (int i = 0; i < pqs.length; i++) {
            if (pqs[i].getPaths().length == 0) {
                File fullPath = pqs[i].getBase().toFile();
                long modTime = getFirstModified(fullPath);
                if (modTime < oldest) {
                    oldest = modTime;
                    if (oldest == 0) {
                        return oldest;
                    }
                }
            } else {
                for (IPath path : pqs[i].getPaths()) {
                    File fullPath = pqs[i].getBase().append(path).toFile();
                    long modTime = getFirstModified(fullPath);
                    if (modTime < oldest) {
                        oldest = modTime;
                        if (oldest == 0) {
                            return oldest;
                        }
                    }
                }
            }
        }
        return oldest;
    }

    private static long getFirstModified(final File dir) {
        if (!dir.isDirectory()) {
            return dir.lastModified();
        }

        long modTime = Long.MAX_VALUE;
        for (File entry : dir.listFiles(FILTER)) {
            modTime = Math.min(modTime, getFirstModified(entry));
        }
        return modTime;
    }

    private long getLastModified(final TopLevelAttribute att,
            final IModelCache ctx, final long threshold) throws CoreException {
        PathGroup[] pqs = att.getPathGroups(ctx, null);
        int idx = pqs.length;
        if (idx == 0) {
            return 0L;
        }

        long latest = 0;
        for (int i = 0; i < pqs.length; i++) {
            if (pqs[i].getPaths().length == 0) {
                File fullPath = pqs[i].getBase().toFile();
                long modTime = getLastModified(fullPath, threshold);
                if (modTime > latest) {
                    latest = modTime;
                }
                if (modTime >= threshold) {
                    return latest;
                }
            } else {
                for (IPath path : pqs[i].getPaths()) {
                    File fullPath = pqs[i].getBase().append(path).toFile();
                    long modTime = getLastModified(fullPath, threshold);
                    if (modTime > latest) {
                        latest = modTime;
                    }
                    if (modTime >= threshold) {
                        return latest;
                    }
                }
            }
        }
        return latest;
    }

    private static long getLastModified(final File dir, final long threshold) {
        if (!dir.isDirectory()) {
            return dir.lastModified();
        }

        long modTime = 0;
        for (File entry : dir.listFiles(FILTER)) {
            modTime = Math.max(modTime, getLastModified(entry, threshold));
            if (modTime >= threshold) {
                break;
            }
        }
        return modTime;
    }
}
