/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2011
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * History
 *   18.10.2010 (meinl): created
 */
package org.knime.build.actors;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.eclipse.buckminster.core.actor.AbstractActor;
import org.eclipse.buckminster.core.actor.IActionContext;
import org.eclipse.buckminster.core.cspec.PathGroup;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * This actor transforms the generated content.xml with the content.xslt. This
 * removed the "Uncategorized" category that is created by Buckminster and adds
 * a new category containing all KNIME source features.
 *
 * @author Thorsten Meinl, University of Konstanz
 */
public class P2SiteModifier extends AbstractActor {
    /**
     * {@inheritDoc}
     */
    @Override
    protected IStatus internalPerform(final IActionContext ctx,
            final IProgressMonitor monitor) throws CoreException {
        for (Map.Entry<String, PathGroup[]> e : ctx.getNamedPathGroupArrays()
                .entrySet()) {
            if ("action.requirements".equals(e.getKey())) {
                for (PathGroup pg : e.getValue()) {
                    IPath base = pg.getBase();
                    IPath contentJar = base.append("content.jar");
                    try {
                        monitor.beginTask("Modifying content.xml", 2);
                        modifyContent(contentJar.toFile(), monitor);
                    } catch (Exception ex) {
                        Status status =
                                new Status(IStatus.ERROR, "org.knime.build",
                                        "Could not modify p2 metadata", ex);
                        throw new CoreException(status);
                    }
                }
            }
        }

        monitor.done();
        return Status.OK_STATUS;
    }

    private static void modifyContent(final File contentJar, final IProgressMonitor monitor)
            throws TransformerException, IOException,
            ParserConfigurationException, SAXException {
        DocumentBuilder parser =
                DocumentBuilderFactory.newInstance().newDocumentBuilder();

        monitor.subTask("Parsing content.xml");
        // parse content.xml from content.jar
        JarFile jf = new JarFile(contentJar);
        InputStream contentStream =
                jf.getInputStream(jf.getEntry("content.xml"));
        Document content = parser.parse(new InputSource(contentStream));
        jf.close();
        monitor.worked(1);
        if (monitor.isCanceled()) {
            return;
        }

        // parse XSLT
        TransformerFactory tFactory = TransformerFactory.newInstance();
        InputStream xsltStream =
                P2SiteModifier.class.getResourceAsStream("content.xslt");
        Templates stylesheet =
                tFactory.newTemplates(new StreamSource(xsltStream));
        Transformer transformer = stylesheet.newTransformer();

        monitor.subTask("Transforming content.xml");
        // transform and write back to content.jar
        File contentXml = new File(contentJar.getParentFile(), "content.jar");
        ZipOutputStream zipOut =
                new ZipOutputStream(new FileOutputStream(contentXml));
        zipOut.putNextEntry(new ZipEntry("content.xml"));

        StreamResult result = new StreamResult(zipOut);
        transformer.transform(new DOMSource(content.getDocumentElement()),
                result);
        zipOut.closeEntry();
        zipOut.close();
        monitor.worked(2);
    }
}
