/* ------------------------------------------------------------------
 * This source code, its documentation and all appendant files
 * are protected by copyright law. All rights reserved.
 *
 * Copyright, 2003 - 2011
 * University of Konstanz, Germany
 * Chair for Bioinformatics and Information Mining (Prof. M. Berthold)
 * and KNIME GmbH, Konstanz, Germany
 *
 * You may not modify, publish, transmit, transfer or sell, reproduce,
 * create derivative works from, distribute, perform, display, or in
 * any way exploit any of the content, in whole or in part, except as
 * otherwise expressly permitted in writing by the copyright owner or
 * as specified in the license file distributed with this product.
 *
 * If you have any questions please contact the copyright holder:
 * website: www.knime.org
 * email: contact@knime.org
 * ---------------------------------------------------------------------
 *
 */
package org.knime.build.tools;

import static org.knime.build.tools.XMLHelper.parser;
import static org.knime.build.tools.XMLHelper.xPathFactory;

import java.io.File;
import java.io.IOException;
import java.util.regex.Pattern;

import javax.xml.transform.TransformerException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.batik.transcoder.TranscoderException;
import org.tmatesoft.svn.core.SVNException;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Checks if plugins have certain necessary prerequisites for building such as
 * an <tt>buckminster.cspex</tt> if they contribute nodes.
 *
 * @author Thorsten Meinl, University of Konstanz
 */
public final class CheckPlugins {
    private CheckPlugins() {
    }

    private static void usage() {
        System.out.print("Usage: " + CheckPlugins.class.getCanonicalName());
        System.out.println(" sourceDirectory [pluginPattern]");
    }

    public static void main(final String[] args) throws Exception {
        if (args.length < 1) {
            usage();
            System.exit(1);
        }
        File sourceDir = new File(args[0]);
        if (!sourceDir.isDirectory()) {
            throw new IllegalArgumentException("Directory '" + sourceDir
                    + "' does not exist");
        }

        Pattern pattern;
        if (args.length > 1) {
            pattern = Pattern.compile(args[1]);
        } else {
            pattern = Pattern.compile(".*");
        }

        if (!recSearch(sourceDir, pattern)) {
            System.exit(1);
        }
    }

    private static boolean recSearch(final File sourceDir, final Pattern pattern)
            throws IOException, TransformerException, SAXException,
            XPathExpressionException, SVNException, TranscoderException {
        boolean ok = true;

        for (File dir : sourceDir.listFiles()) {
            if (dir.isDirectory() && !dir.isHidden()) {
                if (pattern.matcher(dir.getName()).matches()) {
                    File pluginFile = new File(dir, "plugin.xml");
                    if (pluginFile.exists()) {
                        ok &= checkBuckminsterHelpExtension(dir);
                    }
                } else {
                    recSearch(dir, pattern);
                }
            }
        }

        return ok;
    }

    private static boolean checkBuckminsterHelpExtension(final File pluginDir)
            throws SAXException, IOException, XPathExpressionException {
        File pluginFile = new File(pluginDir, "plugin.xml");

        Document doc = parser.parse(pluginFile);

        XPath xpath = xPathFactory.newXPath();
        NodeList nl =
                (NodeList)xpath
                        .evaluate(
                                "/plugin/extension[@point='org.knime.workbench.repository.nodes']/node",
                                doc.getDocumentElement(),
                                XPathConstants.NODESET);
        if (nl.getLength() > 0) { // plugin contributes nodes
            File cspex = new File(pluginDir, "buckminster.cspex");
            if (!cspex.exists()) {
                System.err.println("No buckminster.cspex in '" + pluginDir
                        + "'");
                return false;
            }

            doc = parser.parse(cspex);
            nl = (NodeList)xpath.evaluate("/cspecExtension/alterGroups/" +
                    "private[@name = 'jar.contents']/" +
                    "attribute[@name = 'knime.create.helpfiles']",
                    doc.getDocumentElement(), XPathConstants.NODESET);
            if (nl.getLength() < 1) {
                System.err.println("Help files creating in '" + pluginDir
                        + "' is missing in buckminster.cspex");
                return false;
            }
        }
        return true;
    }
}
