/* ------------------------------------------------------------------
 * This source code, its documentation and all appendant files
 * are protected by copyright law. All rights reserved.
 *
 * Copyright, 2003 - 2011
 * University of Konstanz, Germany
 * Chair for Bioinformatics and Information Mining (Prof. M. Berthold)
 * and KNIME GmbH, Konstanz, Germany
 *
 * You may not modify, publish, transmit, transfer or sell, reproduce,
 * create derivative works from, distribute, perform, display, or in
 * any way exploit any of the content, in whole or in part, except as
 * otherwise expressly permitted in writing by the copyright owner or
 * as specified in the license file distributed with this product.
 *
 * If you have any questions please contact the copyright holder:
 * website: www.knime.org
 * email: contact@knime.org
 * ---------------------------------------------------------------------
 *
 * History
 *   05.08.2010 (meinl): created
 */
package org.knime.build.tools;

import static org.knime.build.tools.XMLHelper.LINE_SEP;
import static org.knime.build.tools.XMLHelper.copy;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.net.URL;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.batik.transcoder.TranscoderException;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.osgi.util.ManifestElement;
import org.eclipse.pde.internal.core.converter.PluginConverter;
import org.eclipse.pde.internal.core.feature.AbstractFeatureModel;
import org.eclipse.pde.internal.core.feature.Feature;
import org.eclipse.pde.internal.core.feature.WorkspaceFeatureModel;
import org.eclipse.pde.internal.core.ifeature.IFeature;
import org.eclipse.pde.internal.core.ifeature.IFeatureURL;
import org.eclipse.pde.internal.core.ifeature.IFeatureURLElement;
import org.eclipse.pde.internal.core.iproduct.IAboutInfo;
import org.eclipse.pde.internal.core.product.ProductModel;
import org.osgi.framework.BundleException;
import org.tmatesoft.svn.core.SVNDepth;
import org.tmatesoft.svn.core.SVNException;
import org.tmatesoft.svn.core.wc.SVNClientManager;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * Program that recursively sets the version in all found plugins and features.
 *
 * @author Thorsten Meinl, University of Konstanz
 *
 */
@SuppressWarnings("restriction")
public final class SetVersion {
    private static final Pattern UPDATE_URL_PATTERN = Pattern
            .compile("/update/\\d+\\.\\d+");

    private SetVersion() {
    }

    private static void usage() {
        System.out.print("Usage: " + SetVersion.class.getCanonicalName());
        System.out.println(" sourceDirectory newVersion [pluginPattern]");
    }

    /**
     * It's a main method...
     *
     * @param args
     * @throws Exception
     */
    public static void main(final String[] args) throws Exception {
        if (args.length < 2) {
            usage();
            System.exit(1);
        }
        File sourceDir = new File(args[0]);
        if (!sourceDir.isDirectory()) {
            throw new IllegalArgumentException("Directory '" + sourceDir
                    + "' does not exist");
        }

        String newVersion = args[1];
        if (!newVersion.matches("^\\d+\\.\\d+\\.\\d+$")) {
            System.err
                    .println("Illegal version format, major.minor.rev expected");
            System.exit(1);
        }
        Pattern pattern;
        if (args.length > 2) {
            pattern = Pattern.compile(args[2]);
        } else {
            pattern = Pattern.compile(".*\\.knime\\..*");
        }

        recSearch(sourceDir, newVersion, pattern);
    }

    private static void recSearch(final File sourceDir,
            final String newVersion, final Pattern pattern) throws IOException,
            SVNException, CoreException, BundleException,
            XPathExpressionException, TranscoderException,
            TransformerException, SAXException {
        for (File dir : sourceDir.listFiles()) {
            if (dir.isDirectory() && !dir.isHidden()) {
                if (pattern.matcher(dir.getName()).matches()) {
                    File manifestFile =
                            new File(dir, "META-INF" + File.separator
                                    + "MANIFEST.MF");
                    if (manifestFile.exists()) {
                        changeBundleManifest(manifestFile, newVersion);
                    }

                    File featureFile = new File(dir, "feature.xml");
                    if (featureFile.exists()) {
                        changeFeature(featureFile, newVersion);
                    }

                    for (File productFile : dir.listFiles()) {
                        if (productFile.isFile()
                                && productFile.getName().endsWith(".product")) {
                            changeProduct(productFile, newVersion);
                        }
                    }

                    File propertiesFile = new File(dir, "plugin.properties");
                    if (propertiesFile.exists()) {
                        changePluginAbout(propertiesFile, newVersion);
                    }

                    File aboutFile = new File(dir, "about.ini");
                    if (aboutFile.exists()) {
                        changeAboutIni(aboutFile, newVersion);
                    }

                    File p2Inf = new File(dir, "p2.inf");
                    if (p2Inf.exists()) {
                        changeP2Inf(p2Inf, newVersion);
                    }

                    File pom = new File(dir, "pom.xml");
                    if (pom.exists()) {
                        changePom(pom, newVersion);
                    }


                    // check log4j config
                    if ("org.knime.core".equals(dir.getName())) {
                        File configDir =
                                new File(dir, "src" + File.separator + "log4j");
                        File current =
                                new File(configDir, "log4j-" + newVersion
                                        + ".xml");
                        if (!current.exists()) {
                            File previous =
                                    getPreviousLog4jFile(configDir, newVersion);
                            System.out.println("Copying previous log4j.xml "
                                    + previous.getName() + " to "
                                    + current.getName());
                            copy(previous, current);
                            SVNClientManager svnManager =
                                    SVNClientManager.newInstance();
                            svnManager.getWCClient().doAdd(current, false,
                                    false, false, SVNDepth.EMPTY, false, false);
                        }
                    }

                    // change version in splash screen
                    if ("org.knime.product".equals(dir.getName())) {
                        File splashFile = new File(dir, "splash.svg");
                        System.out.println("Changing version in "
                                + splashFile.getAbsolutePath());
                        SplashTool.changeSplashVersion(splashFile,
                                splashFile.getParentFile(), newVersion);
                    }
                } else {
                    recSearch(dir, newVersion, pattern);
                }
            }
        }
    }

    private static void changeP2Inf(final File p2Inf, final String newVersion)
            throws IOException {
        Properties props = new Properties();
        Reader in = new BufferedReader(new FileReader(p2Inf));
        props.load(in);
        in.close();

        String configInst = props.getProperty("instructions.configure");
        if (configInst != null) {
            Matcher m = UPDATE_URL_PATTERN.matcher(configInst);
            if (m.find()) {
                String[] newParts = newVersion.split("\\.");
                configInst =
                        m.replaceAll("/update/" + newParts[0] + "."
                                + newParts[1]);
                props.setProperty("instructions.configure", configInst);
                System.out.println("Changing Update Site URLs in " + p2Inf);
            }
        }
    }

    private static void changeAboutIni(final File aboutFile,
            final String newVersion) throws IOException {
        System.out.println("Changing " + aboutFile.getAbsolutePath());
        StringBuilder buf = new StringBuilder();
        BufferedReader in = new BufferedReader(new FileReader(aboutFile));

        String line;
        while ((line = in.readLine()) != null) {
            line =
                    line.replaceAll("Version:\\s+\\d+\\.\\d+\\.\\d+",
                            "Version: " + newVersion);
            buf.append(line).append(LINE_SEP);
        }
        in.close();

        FileWriter out = new FileWriter(aboutFile);
        out.write(buf.toString());
        out.close();
    }

    private static void changePluginAbout(final File propertiesFile,
            final String newVersion) throws IOException, CoreException {
        System.out.println("Changing " + propertiesFile.getAbsolutePath());
        StringBuilder buf = new StringBuilder();
        BufferedReader in = new BufferedReader(new FileReader(propertiesFile));

        String line;
        while ((line = in.readLine()) != null) {
            line =
                    line.replaceAll("Welcome to KNIME\\s+\\d+\\.\\d+\\.\\d+",
                            "Welcome to KNIME " + newVersion);
            buf.append(line).append(LINE_SEP);
        }
        in.close();

        FileWriter out = new FileWriter(propertiesFile);
        out.write(buf.toString());
        out.close();
    }

    private static void changeProduct(final File productFile,
            final String newVersion) throws IOException, CoreException {
        System.out.println("Changing product " + productFile.getAbsolutePath());

        ProductModel productModel = new ProductModel();
        FileInputStream in = new FileInputStream(productFile);
        productModel.load(in, true);
        in.close();

        // change product version
        String version = productModel.getProduct().getVersion();
        String[] parts = version.split("\\.");

        if (parts.length < 3) {
            throw new IllegalArgumentException("Unexpected product version: "
                    + version);
        } else if (parts.length == 3) {
            version = newVersion;
        } else {
            version = newVersion + "." + parts[3];
        }
        productModel.getProduct().setVersion(version);

        // change version in about info
        IAboutInfo aboutInfo = productModel.getProduct().getAboutInfo();
        if (aboutInfo != null) {
            String text = aboutInfo.getText();
            if (text != null) {
                text =
                        text.replaceFirst("KNIME v\\d+\\.\\d+\\.\\d+",
                                "KNIME v" + newVersion);
                aboutInfo.setText(text);
            } else {
                System.err.println("No about message found in " + productFile);
            }
        } else {
            System.err.println("No about message found in " + productFile);
        }

        PrintWriter out = new PrintWriter(productFile);
        out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        productModel.getProduct().write("", out);
        out.close();
    }

    private static void changeFeature(final File featureFile,
            final String newVersion) throws IOException, CoreException {
        System.out.println("Changing feature " + featureFile.getAbsolutePath());

        AbstractFeatureModel featureModel = new WorkspaceFeatureModel();

        FileInputStream in = new FileInputStream(featureFile);
        featureModel.load(in, true);
        in.close();

        IFeature feature = featureModel.getFeature();

        String version = feature.getVersion();
        String[] parts = version.split("\\.");

        if (parts.length < 3) {
            throw new IllegalArgumentException("Unexpected feature version: "
                    + version);
        } else if (parts.length == 3) {
            version = newVersion;
        } else {
            version = newVersion + "." + parts[3];
        }

        feature.setVersion(version);

        // change update site URL
        IFeatureURL url = feature.getURL();
        if (url != null) {
            IFeatureURLElement updateUrl = url.getUpdate();
            if (updateUrl != null) {
                String s = updateUrl.getURL().toExternalForm();

                Matcher m = UPDATE_URL_PATTERN.matcher(s);
                if (m.find()) {
                    String[] newParts = newVersion.split("\\.");
                    s =
                            m.replaceFirst("/update/" + newParts[0] + "."
                                    + newParts[1]);
                    updateUrl.setURL(new URL(s));
                    System.out.println("\tchanging Update Site URL to "
                            + updateUrl.getURL());
                }
            }
        }

        PrintWriter out = new PrintWriter(featureFile);
        out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        ((Feature)feature).write("", out);
        out.close();
    }

    private static final String BUNDLE_VERSION = "Bundle-Version";

    private static void changeBundleManifest(final File manifestFile,
            final String newVersion) throws IOException, BundleException {
        System.out.println("Changing MANIFEST "
                + manifestFile.getAbsolutePath());

        InputStream in = new FileInputStream(manifestFile);
        @SuppressWarnings("unchecked")
        Map<Object, Object> attrs =
                ManifestElement.parseBundleManifest(in, null);
        in.close();

        for (Map.Entry<Object, Object> o : attrs.entrySet()) {
            if (o.getKey().equals(BUNDLE_VERSION)) {
                String version = o.getValue().toString();
                String[] parts = version.split("\\.");

                if (parts.length < 3) {
                    throw new IllegalArgumentException(
                            "Unexpected bundle version: " + version);
                } else if (parts.length == 3) {
                    version = newVersion;
                } else {
                    version = newVersion + "." + parts[3];
                }
                o.setValue(version);
            } else {
                o.setValue(splitOnComma(o.getKey().toString(), o.getValue()
                        .toString()));
            }
        }

        FileWriter writer = new FileWriter(manifestFile);
        new PluginConverter(null).writeManifest(attrs, writer);
        writer.close();
    }

    private static void changePom(final File pomFile, final String newVersion)
            throws IOException, TransformerException, XPathExpressionException,
            SAXException {
        String[] parts = newVersion.split("\\.");
        String version = parts[0] + "." + parts[1];
        String qualifier = "." + parts[2];

        Document doc = XMLHelper.parser.parse(pomFile);
        XPath xpath = XMLHelper.xPathFactory.newXPath();
        Element versionNode =
                (Element)xpath.evaluate("/project/properties/knime.es.version",
                        doc.getDocumentElement(), XPathConstants.NODE);
        if (versionNode != null) {
            System.out.println("Changing pom.xml " + pomFile.getAbsolutePath());
            versionNode.removeChild(versionNode.getFirstChild());
            versionNode.appendChild(doc.createTextNode(version));

            Element qualifierNode =
                    (Element)xpath.evaluate(
                            "/project/properties/knime.es.qualifier",
                            doc.getDocumentElement(), XPathConstants.NODE);
            qualifierNode.removeChild(qualifierNode.getFirstChild());
            qualifierNode.appendChild(doc.createTextNode(qualifier));

            Source domSource = new DOMSource(doc);
            Result streamResult = new StreamResult(pomFile);
            XMLHelper.serializer.transform(domSource, streamResult);
        }
    }

    private static String splitOnComma(final String key, final String value) {
        if ((value.indexOf('\n') >= 0) || ((key + ": " + value).length() <= 78)) {
            return value;
        }
        String[] values = ManifestElement.getArrayFromList(value);
        if (values == null || values.length == 0) {
            return value;
        }
        StringBuffer sb =
                new StringBuffer(value.length() + (values.length - 1)
                        * PluginConverter.LIST_SEPARATOR.length());
        for (int i = 0; i < values.length - 1; i++) {
            sb.append(values[i]).append(PluginConverter.LIST_SEPARATOR);
        }
        sb.append(values[values.length - 1]);
        return sb.toString();
    }

    private static File getPreviousLog4jFile(final File configDir,
            final String newVersion) {
        String[] parts = newVersion.split("\\.");
        int major = Integer.parseInt(parts[0]);
        int minor = Integer.parseInt(parts[1]);
        int revision = Integer.parseInt(parts[2]);

        File prev = null;
        for (int maj = major; (maj >= 1) && (prev == null); maj--) {
            outer: for (int min = 0;; min++) {
                for (int rev = 0;; rev++) {
                    if ((maj == major) && (min == minor) && (rev == revision)) {
                        continue;
                    }

                    File f =
                            new File(configDir + File.separator + "log4j-"
                                    + maj + "." + min + "." + rev + ".xml");
                    if (f.exists()) {
                        prev = f;
                    } else if (rev == 0) {
                        break outer;
                    } else {
                        break;
                    }

                }
            }
        }
        return prev;
    }
}
