/* ------------------------------------------------------------------
 * This source code, its documentation and all appendant files
 * are protected by copyright law. All rights reserved.
 *
 * Copyright, 2003 - 2011
 * University of Konstanz, Germany
 * Chair for Bioinformatics and Information Mining (Prof. M. Berthold)
 * and KNIME GmbH, Konstanz, Germany
 *
 * You may not modify, publish, transmit, transfer or sell, reproduce,
 * create derivative works from, distribute, perform, display, or in
 * any way exploit any of the content, in whole or in part, except as
 * otherwise expressly permitted in writing by the copyright owner or
 * as specified in the license file distributed with this product.
 *
 * If you have any questions please contact the copyright holder:
 * website: www.knime.org
 * email: contact@knime.org
 * ---------------------------------------------------------------------
 *
 * History
 *   06.08.2010 (meinl): created
 */
package org.knime.build.tools;

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.imageio.ImageIO;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.batik.transcoder.TranscoderException;
import org.apache.batik.transcoder.TranscoderInput;
import org.apache.batik.transcoder.TranscoderOutput;
import org.apache.batik.transcoder.XMLAbstractTranscoder;
import org.apache.batik.transcoder.image.PNGTranscoder;
import org.apache.batik.util.XMLResourceDescriptor;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import org.xml.sax.SAXException;

public final class SplashTool {
    private static class MyPNGTranscoder extends PNGTranscoder {
        @Override
        public BufferedImage createImage(final int w, final int h) {
            return new BufferedImage(w, h, BufferedImage.TYPE_INT_BGR);
        }
    }

    private static final String parserClassName;

    static {
        /*
         * sometimes strange error message such as 'SAX2 driver class
         * org.apache.xerces.parsers.SAXParser not found' appear. So we make
         * sure that the XML parser is really found.
         */
        String className =
                "com.sun.org.apache.xerces.internal.parsers.SAXParser";
        // works only in Sun JREs
        try {
            SplashTool.class.getClassLoader().loadClass(className);
        } catch (ClassNotFoundException e) {
            className = null;
        }
        if (className == null) {
            className = XMLResourceDescriptor.getXMLParserClassName();
            try {
                SplashTool.class.getClassLoader().loadClass(className);
            } catch (ClassNotFoundException ex) {
                className = null;
            }
        }

        parserClassName = className;
    }

    private SplashTool() {
    }

    public static void changeSplashVersion(final File splashFile,
            final File destDir, final String newVersion)
            throws TranscoderException, IOException, TransformerException,
            XPathExpressionException, SAXException {
        Document doc = XMLHelper.parser.parse(splashFile);

        XPath xpath = XMLHelper.xPathFactory.newXPath();
        Element versionNode =
                (Element)xpath.evaluate("//text[@id = 'KNIME-Version']/tspan",
                        doc.getDocumentElement(), XPathConstants.NODE);
        Text text = (Text)versionNode.getFirstChild();
        versionNode.removeChild(text);
        versionNode.appendChild(doc.createTextNode("Version " + newVersion));

        File changedFile = new File(destDir, splashFile.getName());
        Source domSource = new DOMSource(doc);
        Result streamResult = new StreamResult(changedFile);
        XMLHelper.serializer.transform(domSource, streamResult);
        convertSvg(changedFile, destDir);
    }

    public static void changeSplashDate(final File splashFile,
            final File destDir, final String date) throws TransformerException,
            IOException, TranscoderException, XPathExpressionException,
            SAXException {
        Document doc = XMLHelper.parser.parse(splashFile);

        XPath xpath = XMLHelper.xPathFactory.newXPath();
        Element builddateNode =
                (Element)xpath.evaluate(
                        "//text[@id = 'KNIME-Builddate']/tspan",
                        doc.getDocumentElement(), XPathConstants.NODE);
        Text text = (Text)builddateNode.getFirstChild();
        builddateNode.removeChild(text);
        builddateNode.appendChild(doc.createTextNode(date));

        File changedFile = new File(destDir, splashFile.getName());
        Source domSource = new DOMSource(doc);
        Result streamResult = new StreamResult(changedFile);
        XMLHelper.serializer.transform(domSource, streamResult);
        convertSvg(changedFile, destDir);
    }

    private static void convertSvg(final File splashFile, final File destDir)
            throws IOException, TranscoderException {
        PNGTranscoder svg2png = new MyPNGTranscoder();
        if (parserClassName != null) {
            svg2png.addTranscodingHint(
                    XMLAbstractTranscoder.KEY_XML_PARSER_CLASSNAME,
                    parserClassName);
        }

        TranscoderInput in = new TranscoderInput(splashFile.toURI().toString());
        File splashPNG =
                new File(destDir, splashFile.getName().replace(".svg", ".png"));
        FileOutputStream os = new FileOutputStream(splashPNG);
        TranscoderOutput out = new TranscoderOutput(os);
        svg2png.transcode(in, out);
        os.close();

        File splashBMP =
                new File(destDir, splashFile.getName().replace(".svg", ".bmp"));

        ImageIO.write(ImageIO.read(splashPNG), "bmp", splashBMP);
    }

    private static final String LINE_SEP = System.getProperty("line.separator");

    private static final Pattern BUILD_DATE_PATTERN = Pattern
            .compile("public static final String BUILD_DATE = \".+\";");

    public static void changeConstantsDate(final File knimeConstants,
            final String date) throws IOException {
        BufferedReader in = new BufferedReader(new FileReader(knimeConstants));
        StringBuilder buf = new StringBuilder((int)knimeConstants.length());
        String line;
        while ((line = in.readLine()) != null) {
            Matcher m = BUILD_DATE_PATTERN.matcher(line);
            if (m.find()) {
                line = m.replaceFirst("public static final String "
                                + "BUILD_DATE = \"" + date + "\";");
            }
            buf.append(line).append(LINE_SEP);
        }
        in.close();

        FileWriter out = new FileWriter(knimeConstants);
        out.write(buf.toString());
        out.close();
    }
}
