/* ------------------------------------------------------------------
 * This source code, its documentation and all appendant files
 * are protected by copyright law. All rights reserved.
 *
 * Copyright, 2003 - 2011
 * University of Konstanz, Germany
 * Chair for Bioinformatics and Information Mining (Prof. M. Berthold)
 * and KNIME GmbH, Konstanz, Germany
 *
 * You may not modify, publish, transmit, transfer or sell, reproduce,
 * create derivative works from, distribute, perform, display, or in
 * any way exploit any of the content, in whole or in part, except as
 * otherwise expressly permitted in writing by the copyright owner or
 * as specified in the license file distributed with this product.
 *
 * If you have any questions please contact the copyright holder:
 * website: www.knime.org
 * email: contact@knime.org
 * ---------------------------------------------------------------------
 *
 */
package org.knime.build.tools;

import static org.knime.build.tools.XMLHelper.LINE_SEP;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.transform.TransformerException;
import javax.xml.xpath.XPathExpressionException;

import org.apache.batik.transcoder.TranscoderException;
import org.tmatesoft.svn.core.SVNException;
import org.xml.sax.SAXException;

public final class UpdateCopyrightYear {
    private UpdateCopyrightYear() {

    }

    // matches Copyright (C) 2003 - 2011 and also
    // Copyright, 2003 - ....
    private static final Pattern COPYRIGHT = Pattern
            .compile("(Copyright,? (?:\\(C\\) )?20\\d{2,2} - )(20\\d{2,2})");


    private static final Pattern EXCLUDE_FILES = Pattern.compile(".+\\.(?:jar|zip|png|gif|bmp|dll)$");

    private static void usage() {
        System.out.print("Usage: " + UpdateCopyrightYear.class.getCanonicalName());
        System.out.println(" sourceDirectory year [pluginPattern]");
    }

    /**
     * @param args
     */
    public static void main(final String[] args) throws Exception {
        if (args.length < 2) {
            usage();
            System.exit(1);
        }
        File sourceDir = new File(args[0]);
        if (!sourceDir.isDirectory()) {
            throw new IllegalArgumentException("Directory '" + sourceDir
                    + "' does not exist");
        }

        String newVersion = args[1];
        Pattern pattern;
        if (args.length > 2) {
            pattern = Pattern.compile(args[2]);
        } else {
            pattern = Pattern.compile(".*");
        }

        recSearch(sourceDir, newVersion, pattern);
    }

    private static void recSearch(final File sourceDir,
            final String year, final Pattern pattern) throws IOException,
            TransformerException, SAXException, XPathExpressionException,
            SVNException, TranscoderException {
        for (File dir : sourceDir.listFiles()) {
            if (dir.isDirectory() && !dir.isHidden()) {
                if (pattern.matcher(dir.getName()).matches()) {
                    updateFiles(dir, year);
                } else {
                    recSearch(dir, year, pattern);
                }
            }
        }
    }

    private static void updateFiles(final File sourceDir,
            final String year) throws IOException,
            TransformerException, SAXException, XPathExpressionException,
            SVNException, TranscoderException {
        for (File f : sourceDir.listFiles()) {
            if (f.isDirectory() && !f.isHidden() && !".svn".equals(f.getName())) {
                updateFiles(f, year);
            } else if (f.isFile() && !EXCLUDE_FILES.matcher(f.getName()).matches()) {
                updateCopyrightYear(f, year);
            }
        }
    }

    private static void updateCopyrightYear(final File file, final String year)
            throws IOException {
        StringBuilder buf = new StringBuilder();
        BufferedReader in = new BufferedReader(new FileReader(file));
        String line;
        boolean match = false;
        while ((line = in.readLine()) != null) {
            Matcher m = COPYRIGHT.matcher(line);
            if (m.find() && !year.equals(m.group(2))) {
                match = true;
                buf.append(m.replaceAll(m.group(1) + year));
            } else {
                buf.append(line);
            }
            buf.append(LINE_SEP);
        }
        in.close();

        if (match) {
            System.out.println("Changing copyright year in " + file);
            FileWriter out = new FileWriter(file);
            out.write(buf.toString());
            out.close();
        }

    }
}
